/* 
   alsa_core.c (C) Heikki Lindholm 2003 
   
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

/*
 * NOTE: no mixer settings in this one, because there's already the 
 * command line 'amixer' to do that 
 * ALSO: alsa core reports audio drops wheras OSS core doesn't and might
 *       therefore seem to work better, but when Warp increases steadily
 *       while capturing with OSS core it means it's dropping audio
 *
 * TODO: put also alloc_frags part to audio_common.c	
 */

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <sys/ioctl.h>
#include <pthread.h>
#include <signal.h>

#include "audio_core_common.h" // includes nvrec.h

#ifdef HAVE_ALSA

// this is already in two places. Should be put to nvrec.h..
#define TIMESTAMP_TO_NTIME_T(x) \
        (((ntime_t)x.tv_sec * 1000000000ULL) + ((ntime_t)x.tv_usec * 1000ULL))

// read chunk_size frames 
// function copy'n'pasted from aplay.c 
ssize_t do_pcm_read(snd_pcm_t *handle, unsigned char *data, snd_pcm_uframes_t chunk_size,
			unsigned int bits_per_frame) {
	ssize_t r;
	size_t result = 0;
	size_t count;

//	IPRINT("reading: %u frames\n", chunk_size);
	count = chunk_size;

	while (count > 0) {
		r = snd_pcm_readi(handle, data, count);
		if (r == -EAGAIN || (r >= 0 && r < count)) {
			snd_pcm_wait(handle, 1000);
		} else if (r == -EPIPE) {
			snd_pcm_status_t *status;
			int res;
        
			IPRINT("\nsuspected audio xrun\n");
			snd_pcm_status_alloca(&status);
			if ((res = snd_pcm_status(handle, status))<0) {
				EPRINT("status error: %s\n", snd_strerror(res));
				break;
			}
			if (snd_pcm_status_get_state(status) == SND_PCM_STATE_XRUN) {
				struct timeval tstamp;
				ntime_t now, diff;
				now = getntime();
				snd_pcm_status_get_trigger_tstamp(status, &tstamp);
				diff = now - TIMESTAMP_TO_NTIME_T(tstamp);
				IPRINT("\naudio dropped for at least %lld ns long!\n", diff);
				if ((res = snd_pcm_prepare(handle))<0) {
					EPRINT("xrun: prepare error: %s\n",
							snd_strerror(res));
					break;
				}
			}
		} else if (r == -ESTRPIPE) {
			int res;
			IPRINT("suspended while reading.\n");
			while ((res = snd_pcm_resume(handle)) == -EAGAIN)
				sleep(1);
			if (res < 0) {
				EPRINT("resume failed. restarting...\n");
				if ((res = snd_pcm_prepare(handle)) < 0) {
					EPRINT("prep. error: %s\n", snd_strerror(res));
					break;
				}
			}
		} else if (r < 0) {
			EPRINT("read error: %s\n", snd_strerror(r));
			return 0;
		}
		if (r > 0) {
			result += r;
			count -= r;
			data += r * bits_per_frame / 8;
		}
	}
	return chunk_size;
}

void * alsa_core_thread(void * t)
{
	volatile struct audio_core_common_t * acomt = (struct audio_core_common_t *)t;
	int i;
	ntime_t ntime;
//	unsigned long long u;
	struct sched_param sp;
	sigset_t newmask;
	snd_timestamp_t tstamp;
	//snd_pcm_uframes_t ucount;
	snd_pcm_sframes_t delay = 0;
	snd_pcm_status_t *status;
	snd_pcm_status_alloca(&status);
	
	sigfillset(&newmask);
	if(pthread_sigmask(SIG_BLOCK, &newmask, NULL) != 0) {
		PERROR("Could not block signals, but continuing anyway");
	}
	sp.sched_priority = sched_get_priority_max(SCHED_FIFO);
	if(pthread_setschedparam(pthread_self(), SCHED_FIFO, &sp) != 0) {
		sp.sched_priority = sched_get_priority_max(SCHED_RR);
		if(pthread_setschedparam(pthread_self(), SCHED_RR, &sp) != 0) {
			sp.sched_priority = sched_get_priority_max(SCHED_OTHER);
			if(pthread_setschedparam(pthread_self(), SCHED_OTHER, &sp) != 0) {
				PERROR("Could not get any realtime priority");
				EPRINT("\t- This is not a serious problem, but capture quality under high\n");
				EPRINT("\t  load could be improved by running as root.\n");
			}
		}
	}

	IPRINT("Starting!\n");
	if (snd_pcm_start(acomt->handle) < 0)
       		PERROR("start failed");
	
	while(!acomt->exit) {
		if ((acomt->frag + acomt->head)->full == 1) {
			EPRINT("Buffer full - discarding data\n");
			usleep(10000);
			continue;
		}
		i = 0;
		if (do_pcm_read(acomt->handle, (acomt->frag + acomt->head)->buf,
			(acomt->fragsize*8)/acomt->bits_per_frame, acomt->bits_per_frame) == 0) {

			EPRINT("Read failed.\n");
			continue;
		}
		if (snd_pcm_status(acomt->handle, status) < 0) {
			EPRINT("Failed to get status, can't get good timestamp\n");
			ntime = getntime(); // fallback option
		}
		else {
			snd_pcm_status_get_tstamp(status, &tstamp);
			//ucount = snd_pcm_status_get_avail(status);
			delay = snd_pcm_status_get_delay(status);
			//ucount = snd_pcm_status_get_overrange(status);
			ntime = (ntime_t)(1000L*
			((ntime_t)tstamp.tv_sec*1000L*1000L+tstamp.tv_usec));
			ntime -= (((ntime_t)delay*1000L*1000L*1000L)/acomt->real_rate);
		}
		
		i = acomt->head;
		i++;
		if(i >= acomt->nfrags) i = 0;
		(acomt->frag + acomt->head)->full = 1;
		(acomt->frag + acomt->head)->time = ntime;
		acomt->head = i;
	}
	DXPRINT("main loop done\n");
	IPRINT("Exiting!\n");
	DXPRINT("signal main loop done\n");
	acomt->done = 1;
	DXPRINT("leave\n");
	return NULL;
}

void alsa_core_close(struct aud_core_t * audt)
{
	volatile struct audio_core_common_t * acomt = (struct audio_core_common_t *)audt->priv;
	
	DXPRINT("enter\n");
	if(!acomt)
		return;
	DXPRINT("signal alsa core exit\n");
	acomt->exit = 1;
	DXPRINT("wait for alsa core exit\n");
	while(!acomt->done) // FIXME - time this out!
		usleep(100000);
	DXPRINT("alsa core exited\n");
	IPRINT("Read thread terminated.\n");
	DXPRINT("close alsa pcm device\n");
	snd_pcm_close(acomt->handle);
	IPRINT("Freed buffers.\n");
	DXPRINT("free alsa frags\n");
	audio_core_freefrags(acomt);
	IPRINT("Freed fragments.\n");
	DXPRINT("free priv\n");
	free(audt->priv);
	DXPRINT("free audt\n");
	free(audt);
	DXPRINT("leave\n");
	return;
}

void alsa_core_usage()
{
	MPRINT("ALSA capture core usage:\n");
	MPRINT("\t-d <pcm device>                      [default]\n");
	MPRINT("\t-r <audio sample rate>               [44100]\n");
	MPRINT("\t-b <bits per sample>                 [16]\n");
	MPRINT("\t-s stereo                            [mono]\n");
	MPRINT("\t-ps <period size>                    [autodetect]\n");
}

struct aud_core_t alsa_core;

struct aud_core_t * alsa_core_init(char ** args, ntime_t ftime)
{
	struct aud_core_t * act;
	struct audio_core_common_t * acomt;
 	int i, j, t;
 	unsigned long long bpf;
 	pthread_t athread;
 	char * fdev = NULL;
 	int N = -1;
	snd_pcm_hw_params_t *params;
	snd_pcm_sw_params_t *swparams;
	snd_pcm_format_t sample_format;
	snd_pcm_uframes_t buffer_size;
	snd_pcm_uframes_t period_size, period_size_min, period_size_max, forced_period_size = 0;
	unsigned int max_periods;
	
 	NV_ARGS;

	act = malloc(sizeof(struct aud_core_t));
	if(!act) {
		PERROR("Could not malloc new audio core");
		return NULL;
	}
	acomt = malloc(sizeof(struct audio_core_common_t));
	if(!acomt) {
		PERROR("Could not malloc new alsa core");
		free(act);
		return NULL;
	}
	memcpy(act, &alsa_core, sizeof(struct aud_core_t));
	act->priv = acomt;
	memset(acomt, 0, sizeof(struct audio_core_common_t));
 	
	NV_SCAN_ARGS {
		if(strcmp(NV_THIS_ARG, "-r") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -r.\n");
				goto usage;
			}
			acomt->rate = atoi(NV_THIS_ARG);
			if(acomt->rate < 8000) {
				EPRINT("Invalid argument to -r (>=8000).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-b") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -b.\n");
				goto usage;
			}
			acomt->bits = atoi(NV_THIS_ARG);
			if((acomt->bits != 8) && (acomt->bits != 16)) {
				EPRINT("Invalid argument to -b (8 or 16).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-s") == 0) {
			acomt->chans = 2;
		} else if(strcmp(NV_THIS_ARG, "-m") == 0) {
			acomt->chans = 1;
		} else if(strcmp(NV_THIS_ARG, "-d") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -d.\n");
				goto usage;
			}
			fdev = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-ps") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -ps.\n");
				goto usage;
			}
			if (atoi(NV_THIS_ARG) > 0)
				forced_period_size = atoi(NV_THIS_ARG);	
			else {
				EPRINT("Invalid argument to -ps (>0).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-N") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -N.\n");
				goto usage;
			}
			N = atoi(NV_THIS_ARG);
			if(N < 4) {
				EPRINT("Invalid argument to -N (>=4).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-vr") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vr.\n");
				goto usage;
			}
			ftime = (unsigned long long)(1e9 / atof(NV_THIS_ARG));
		}
	}
	IPRINT("%d bps, %d bits, %s, from %s\n", acomt->rate, acomt->bits, (acomt->chans == 1)?"mono":"stereo", fdev);

 	bpf = acomt->rate * acomt->chans * ((acomt->bits==16)?2:1);
 	bpf *= ftime;
 	bpf /= 1000000000LL;
 	t = bpf / 4;
	IPRINT("bpf: %llu t: %u\n", bpf, t);
 	
	if (snd_pcm_open(&acomt->handle, "default", SND_PCM_STREAM_CAPTURE, 0) < 0) {
		PERROR("Open ALSA device");
		goto error0;
	}
  
	snd_pcm_hw_params_alloca(&params);
	snd_pcm_sw_params_alloca(&swparams);
	if (snd_pcm_hw_params_any(acomt->handle, params) < 0) {
		PERROR("Get any hw parameters failed");
		goto error1;
	}
	if (snd_pcm_hw_params_set_access(acomt->handle, 
						params, SND_PCM_ACCESS_RW_INTERLEAVED) < 0) {
		PERROR("Set interleaved access type failed");
		goto error1;
        }

	// map sample format
	if (acomt->bits == 8)
		sample_format = SND_PCM_FORMAT_S8;
	else if (acomt->bits == 16)
		sample_format = SND_PCM_FORMAT_S16_LE;
	else {
		EPRINT("Unimplemented sample format mapping!\n");
		goto error1;
	}
	if (snd_pcm_hw_params_set_format(acomt->handle, params, sample_format) < 0) {
                PERROR("Could not set sample format");
		goto error1;
        }
	if (snd_pcm_hw_params_set_channels(acomt->handle, params, acomt->chans) < 0) {
		PERROR("Could not set no. channels");
		goto error1;
	}
	j = acomt->rate;
	if (snd_pcm_hw_params_set_rate_near(acomt->handle, params, &j, 0) < 0) {
		PERROR("Could not set sampling rate");
		goto error1;
	}
	acomt->real_rate = j;
 	acomt->bps = acomt->rate * acomt->chans * ((acomt->bits==16)?2:1);
 	acomt->realbps = j * acomt->chans * ((acomt->bits==16)?2:1);
	IPRINT("bps=%d, realbps=%d\n", acomt->bps, acomt->realbps);

	acomt->bits_per_frame = snd_pcm_format_physical_width(sample_format) * acomt->chans;
	IPRINT("bits_per_frame = %u\n", acomt->bits_per_frame);

	//j = t/1000; // period time is in microsecs
	// save max period size
	if (snd_pcm_hw_params_get_periods_max(params, &max_periods, &j) < 0) {
		PERROR("Could not get max no. periods");
		goto error1;
	}
	if (snd_pcm_hw_params_get_period_size_min(params, &period_size_min, &j) < 0) {
		PERROR("Could not get minimum period size");
		goto error1;
	}
	// if the exact value is larger, integerize the minimum
	if (j > 0)
		period_size_min++; 
	if (snd_pcm_hw_params_get_period_size_max(params, &period_size_max, &j) < 0) {
		PERROR("Could not get maximum period size");
		goto error1;
	}
	// if the exact value is smaller, integerize the maximum
	if (j > 0)
		period_size_max--;
	// try to find out bigger period sizes that are multiples of the wanted one
	period_size = 0;
	for (j = 2; ; j++) {
		ntime_t	num, den;
		snd_pcm_uframes_t per;
		num = acomt->real_rate * ftime * j;
		den = 1000000000LL;
		if (num / den > (ntime_t)period_size_max) 
			break;
		if (num % den != 0) // try only integer sizes, except the first original one
			continue;
		per = (snd_pcm_uframes_t)(num/den);
		
		if (snd_pcm_hw_params_test_period_size(acomt->handle, params, per, 0) == 0) {
			IPRINT("sensible period size(1): %u\n", (unsigned int)per);
			period_size = per;
		}
	}
	// go through possible periods that could be fractions of the wanted size
	for (j = 1; ; j++) {
		ntime_t	num, den;
		snd_pcm_uframes_t per;
		num = acomt->real_rate * ftime;
		den = 1000000000LL*j;
		if (num / den < (ntime_t)period_size_min) 
			break;
		if (num % den != 0 && j != 1) // try only integer sizes, except the first original one
			continue;
		per = (snd_pcm_uframes_t)(num/den);
		
		if (snd_pcm_hw_params_test_period_size(acomt->handle, params, per, 0) == 0) {
			IPRINT("sensible period size(2): %u\n", (unsigned int)per);
			if (period_size == 0) // take the largest
				period_size = per;
		}
	}
	// no matches found -> use frame time
	if (period_size == 0) 
		period_size = acomt->real_rate / (1000000000LL / ftime);
	
	IPRINT("max periods: %u min size: %u max size: %u\n", max_periods,
				(unsigned int)period_size_min, (unsigned int)period_size_max);
	
	//buffer_size = max_periods*period_size;
	if (forced_period_size > 0) {
		IPRINT("using period size from command line\n");
		period_size = forced_period_size;
	}
	IPRINT("requesting period size %u\n", (unsigned int)period_size);
	// first restrict alsa configuration space
	forced_period_size = period_size; // save the original for set_period_size_near
	if (snd_pcm_hw_params_set_period_size_max(acomt->handle, params, &period_size, 0) < 0) {
		PERROR("Could not set max period size");
		goto error1;
	}
	// then try to set the period
	if (snd_pcm_hw_params_set_period_size_near(acomt->handle, params, &forced_period_size, 0) < 0) {
		PERROR("Could not set period size");
		goto error1;
	}
	// try 1 sec buffer altough it probably won't come close 
	j = 1000000;
	if (snd_pcm_hw_params_set_buffer_time_near(acomt->handle, params, &j, 0) < 0) {
		PERROR("Could not set buffer time");
		goto error1;
	}
	
	// actually set parameters
	if (snd_pcm_hw_params(acomt->handle, params) < 0) {
		PERROR("Could not set HW params");
		goto error1;
	}

	// getting these again here probably isn't required (depends on whether the above can change these)
	snd_pcm_hw_params_get_period_size(params, &period_size, 0);
	snd_pcm_hw_params_get_buffer_size(params, &buffer_size);
	// only one period won't do
	if (period_size == buffer_size) {
		EPRINT("got only one period worth of buffer.\n");
		goto error1;
	}
	IPRINT("period size: %u buffer size: %u\n", (unsigned int)period_size, (unsigned int)buffer_size);
	
	snd_pcm_sw_params_current(acomt->handle, swparams);
	if (snd_pcm_sw_params_set_tstamp_mode(acomt->handle, swparams, SND_PCM_TSTAMP_MMAP) < 0) {
		PERROR("Could not set timestamp mode");
		goto error1;
	}
	if (snd_pcm_sw_params(acomt->handle, swparams) < 0) {
		PERROR("Could not set sw params");
		goto error1;
	}

	if (snd_pcm_prepare(acomt->handle) < 0) { // necessary at all?!?
		PERROR("prepare failed");
		goto error1;
	}

	// try to find a fraction of period for reading more frequently
	i = 2;
	while (period_size > (bpf*8)/acomt->bits_per_frame) {
		if (period_size % i == 0)
			period_size /= i;
		if (period_size/i < (bpf*8)/acomt->bits_per_frame && i > 10) // don't go on forever
			break;
		i++;
	}
		
	// bytes per frag
	j = (period_size*acomt->bits_per_frame)/8;
	acomt->period_size = period_size;
	acomt->fragsize = j;
	i = (N * bpf * 2) / j;
	IPRINT("%i buffer frags required of %i bytes (%i frames)\n", i, j, (j*8)/acomt->bits_per_frame);
	acomt->frag = (struct frag_t *)malloc(sizeof(struct frag_t) * i);
	if(!acomt->frag) {
		PERROR("Could not allocate fragment memory");
		goto error1;
	}
	memset(acomt->frag, 0, sizeof(struct frag_t) * i);
	while(acomt->nfrags < i) {
		(acomt->frag + acomt->nfrags)->buf = (unsigned char *)malloc(acomt->fragsize);
		if(!(acomt->frag + acomt->nfrags)->buf) {
			if(acomt->nfrags > 0) {
				PERROR("Could not allocate memory for all buffers, trying to continue anyway!");
			} else {
				PERROR("Could not allocate memory for any buffers - aborting!");
				goto error1;
			}
		}
		(acomt->frag + acomt->nfrags)->time = -1;
		(acomt->frag + acomt->nfrags)->full = 0;
		acomt->nfrags++;
	}
	acomt->head = 0;
	acomt->tail = 0;
	acomt->exit = 0;
	acomt->done = 0;

	if(pthread_create(&athread, NULL, alsa_core_thread, (void *)acomt) != 0) {
		PERROR("Could not start alsaio thread");
		goto error1;
	}
	return act;

error1:
	audio_core_freefrags(acomt);
	snd_pcm_close(acomt->handle);
error0:
	free(act);
	free(acomt);
	return NULL;
usage:
	EPRINT("Error parsing options.  Please run without any command line options for\n");
	EPRINT("detailed usage information.\n");
	free(act);
	free(acomt);
	return NULL;
}

struct aud_core_t alsa_core = {
	aud_core_init: alsa_core_init, 
	aud_core_close: alsa_core_close,
	aud_core_tailtime: audio_core_tailtime,
	aud_core_droptail: audio_core_droptail,
	aud_core_read: audio_core_read,
	aud_core_bps: audio_core_bps,
	aud_core_overrides: NULL,
	aud_core_usage: alsa_core_usage,
	aud_core_has_data: audio_core_has_data,
	priv: NULL
};

#else

struct aud_core_t * alsa_null_core_init(char ** args, ntime_t ftime)
{
	return NULL;
}

void alsa_null_core_usage()
{
}

struct aud_core_t alsa_core = {
	aud_core_init: alsa_null_core_init,
	aud_core_close: NULL,           
	aud_core_tailtime: NULL,               
	aud_core_droptail: NULL,               
	aud_core_read: NULL,           
	aud_core_bps: NULL,            
	aud_core_overrides: NULL,
	aud_core_usage: alsa_null_core_usage,
	aud_core_has_data: NULL,
	priv: NULL
};

#endif
